/***
 * Simple text to html converter
 *
 * Scans a text file and looks for headings.
 * These headings are used to make a table of contents 
 * with links to the sections.
 *
 * A heading is anyline that starts with a number(s) and is 
 * followed by a blank line.
 *  
 * The idea being that any manual or book usualy contains numbered
 * headings and subheadings. By outputting the doc as straight
 * text, a simple Hypertext representation can be created with
 * this converter.
 * 
 *
 * ( Not really upto distribution standard, not much tested )
 *
 * harry@swin.edu.au  Oct-94
 ***/

#include <stdio.h>
#include <string.h>

extern char *optarg;
extern int optind, opterr;

#define LEN	256	/* line & string lengths */
#define XTRA	5	/* xtra chars that maybe appended to a string */

#define CFGFILE	"/home/staff/harry/txtx2html/txt2html.cfg"

char *WELCOME, *COPYRIGHT, *CRFILE;

#define MYTAG	"Created by txt2html"
#define MYLINK	"/cwis/OPAX/t2h_man/txt2html_0.html"

/* Some common HTML tags */

char *RULE;
char *BODY, *EBODY;
char *HEAD, *EHEAD;
char *TITLE, *ETITLE;
char *HEAD1, *EHEAD1;
char *HEAD2, *EHEAD2;
char *MARK, *EMARK;
char *PARAGRAPH;
char *INDENT, *EINDENT;
char *BREAK;

/* images to use for icons in navigation */
char *NEXT;
char *PREVIOUS;
char *TOP;
char *HOME;

/* text string alternates for above images */
char *TNEXT;
char *TPREVIOUS;
char *TTOP;
char *THOME;

/* chars to escape (have special meaning in html ) */
#define GT	"&gt;"	
#define LT	"&lt;"	
#define AMP	"&amp;"	

struct ENTRY {
	char *var;
	char *val;
	struct ENTRY *next;
};

struct NODE {
	int depth;		/* header depth in TOC */
	char *header;		/* Header string */
	char *linkstring;
	char *filename;		/* name of file containing section */
	int part;		/* page number for this section */ 
	struct NODE *next, *previous;
};

int head_count=0;		/* count headers in pre-scan */
char *filename;			/* base filename of out files */
int file_count=1;		/* Page counter & part of filename */
				/* 0 is for TOC */
static int end_toc=0;
char *title, *backlink;
int gentoc=0;

struct ENTRY *readcfg();
char *get_cfg();

struct NODE *node, *head, *tail;

struct NODE * parse_head();
FILE * open_output();
char *textlink();
char *imagelink();
char *fix_line();
struct NODE *add_node();
char *filter();

main(argc, argv)
int argc;
char *argv[];
{
	long linecount=0;		
	int i=0, j=0, fflag=0, tflag=0, pflag=0,
		keep_format=0;
	FILE *f1, *f2=NULL;		/* text file to translate */
	char *line, inline[LEN], *p, c;

	opterr=0;
	while ((c=getopt(argc, argv, "f:t:p:gk"))!=-1) 
	   switch(c) {
		case 'f' : filename = strdup(optarg);
			   fflag=1;
			   break;
		case 't' : title = strdup(optarg);
			   tflag=1;
			   break;
		case 'p' : backlink = strdup( optarg);
			   pflag=1;
			   break;
		case 'g' : gentoc = 1;
			   break;
		case 'k' : keep_format=1;
			   break;
		case '?' : usage(argv[0]);
	   }

	init_vars();

	if ((f1 = fopen(filename,"r"))==NULL) {
		fprintf(stderr,"Error opening file: %s\n",filename);
		usage(argv[0]);
	}

	if (!backlink)
		backlink=strdup("previous.html");
	if (!title)
		title=strdup(argv[0]);
	
	scan_headers(filename);

	p = (char *) strrchr(filename,'.');
	if (p)
		*p=NULL;

	head=node=NULL;

/* read file line by line */	

	while (fgets(inline,LEN,f1)) {
		linecount++;

		line=filter(inline);

		if (gentoc) {
		   if (!head) { 
	 	      if (head = parse_head(line)) {
			head->previous = NULL;
			head->next = NULL;
			node = head;
			f2 = open_output(f2, node);
			file_count++;
		 	write_headers(f2,node);
		      }
		    
	            } else if (node->next = parse_head(line)) {
			node->next->previous=node;
			write_trailers(f2,node);
			f2 = open_output(f2, node->next);
			file_count++;
			write_headers(f2,node->next);
			node=node->next;
			node->next=NULL;
		   } else {
		      if (keep_format)
		      	strcat(line,BREAK);
		      fprintf(f2,"%s",line);
		   }
		} else { /* gentoc */
		      if ((!check_paragraph(fix_line(line))) && !end_toc) {
		         if (!head) { 
		            if (head=add_node(fix_line(line))) {
			       head->previous = NULL;
			       node=head;
			       node->next=NULL;
		            }
		        } else if (node->next=add_node(fix_line(line))) {
			    node->next->previous=node;
			    node=node->next;
			    node->next=NULL;
		        }
		       } else if (check_paragraph(fix_line(line))) {
			       if (f2)
			           write_trailers(f2, node->previous);
			       f2=open_output(f2,node);
			       write_headers(f2,node);
			       if (node->next)
			           node=node->next;
		       } else {
		            if (keep_format)
		      	        strcat(line,BREAK);
		            fprintf(f2,"%s",line);
		       }
		}		
		i=0;
		while(isspace(line[i])) i++;
		if (i==strlen(line) && f2)
			write_paragraph(f2);
		
		free(line);
	}	

	if (f2)
		write_trailers(f2,node);

	tail=node;

	if (f2)
		fclose(f2);

	write_toc();
}

/***
 * scan_headers()
 *	Scans the initial file for headers so that I know how
 *	files there will be.
 *
 ***/
scan_headers(fname)
char *fname;
{
	FILE *f;
	char line[LEN];
	int depth,i,num;

	if  (f=fopen(fname,"r")) {
		while (fgets(line,LEN,f)) {
		  depth=0; i=0; num=0;
	   	  while(isspace(line[i])) i++;
		  for (;;) {
	   		if (isdigit(line[i])) {
			    if (!num)
			        depth++;	
			    num=1;
	   		} else if (line[i]=='.')
         	   		num=0; 
	   		else { 
				num=0;
				break;
			}
	   		i++;
		  }
		if (depth)
		    head_count++;
		}

	   fclose(f);
	   return(1);
	} else
		return(0);
}

/**
 * parse_head()
 *	
 * 	Looks for a line starting with a Number. It assumes
 *	that these are paragraph titles.
 *	When found these are save in a list for later use
 *	in creating the Table of Contents.
 *
 ***/
struct NODE * parse_head(line)
char *line;		/* pointer to current line */
{
	char *p;
	int i=0, num=0, depth=0,dot=0;
	struct NODE *lnode;

	while(isspace(line[i])) i++;
	for (;;) {
	   if (isdigit(line[i])) {
		if (!num)
		    depth++;	
		num=1; dot=0;
	   } else if (line[i]=='.')
         	   num=0; 
	   else 
		break;
	   i++;

	} /* for */

/* if depth, then we have a heading */

	if (depth) {
	    lnode = (struct NODE *) malloc(sizeof(struct NODE));
	    lnode->depth=depth;

	    lnode->header = (char *) malloc(strlen(line) +1);
	    strcpy(lnode->header, line);
	    if (p=strrchr(lnode->header,'\n'))
		*p=NULL;
		
	    lnode->filename = (char *) malloc(strlen(filename)+10); 		
	    sprintf(lnode->filename,"%s_%d.html",filename,file_count);

	    lnode->part=1;
	    lnode->next=NULL;
	
	    return(lnode);
	} else
	    return(NULL);
}

/***
 * open_output() - Open a file and close the previous 
 *		   one.
 *
 ***/
FILE * open_output(f,lnode)
FILE *f;
struct NODE *lnode;
{
	FILE *f2;

	if (f) fclose(f);

	if (f2 = fopen(lnode->filename,"w")) 
		return(f2);
	else
	    	return(NULL);
		
}

/***
 * write_headers()
 * 	Output any special formatting bits and the header line
 * 	for the section.
 *
 ***/
write_headers(f,lnode)
FILE *f;
struct NODE *lnode;
{
	char fname[256];

	fprintf(f,"<HTML>\n");		/* A html file sig */
	fprintf(f,"%s\n",HEAD);
	fprintf(f,"%s%s%s\n",TITLE,lnode->header,ETITLE);
	fprintf(f,"%s\n",EHEAD);
	fprintf(f,"%s\n",BODY);

	write_navigators(f,lnode);

	fprintf(f,"%s%s%s\n",HEAD2,lnode->header,EHEAD2);
	fprintf(f,"%s\n",RULE);
	return;
}

write_trailers(f,lnode)
FILE *f;
struct NODE *lnode;
{
	fprintf(f,"%s\n",RULE);

	write_navigators(f,lnode);

	fprintf(f,"%s%s%s\n",PARAGRAPH,textlink(CRFILE,COPYRIGHT),PARAGRAPH);
	fprintf(f,"%s\n",textlink(MYLINK, MYTAG));
	fprintf(f,"%s\n",EBODY);

	return;
}

/***
 * write_navigators()
 * 	Write the Navigation links to other pages.
 *
 ***/
write_navigators(f,lnode)
FILE *f;
struct NODE *lnode;
{
	char fname[256];

	if (!lnode->previous) {
/*	if (file_count == 2) { 
*/
		sprintf(fname,"%s_0.html",filename);
		fprintf(f,"%s",imagelink(fname,
			  PREVIOUS, "BOTTOM", TPREVIOUS));

	} else if (lnode->previous) {
		fprintf(f,"%s",imagelink(lnode->previous->filename,
			  PREVIOUS, "BOTTOM", TPREVIOUS));
	}
	if (gentoc) {
	    if (file_count <= head_count) {
		sprintf(fname,"%s_%d.html",filename,file_count);
	        fprintf(f,"%s",imagelink(fname, NEXT, "BOTTOM", TNEXT));	
	    }
	} else if (lnode->next) {
	    fprintf(f,"%s",imagelink(lnode->next->filename, NEXT, 
		    "BOTTOM", TNEXT));	
	}
	sprintf(fname,"%s_0.html",filename);
	fprintf(f,"%s",imagelink(fname,TOP,"BOTTOM",TTOP));

	fprintf(f,"%s",imagelink(WELCOME,HOME,"BOTTOM",THOME));
}

write_paragraph(f)
FILE *f;
{
	fprintf(f,"%s\n",PARAGRAPH);
	return;
}

/***
 * textlink()
 *	return  a link for string to fname
 *
 ***/
char *textlink(fname, string)
char *fname, *string;
{
	static char str[256];

	sprintf(str,"<A HREF=\"%s\"> %s %s",fname,string,EMARK);
	return(str);
}

/***
 * imagelink()
 *	Return line for an image link	
 *
 ***/
char *imagelink(fname, image,align,alt)
char *fname,		/* filename to link to */
     *image,		/* image file */
     *align,		/* alignment */
     *alt;		/* alternate text string for image */
{
	static char str[256];

	sprintf(str,"<A HREF=\"%s\"><IMG SRC=\"%s\" ALT=\"%s \" ALIGN=\"%s\">%s",
		fname,image,alt,align,EMARK);
	return(str);
}
/***
 * write_toc()
 *	Write the table of contents.
 ***/
write_toc()
{
	FILE *f1;
	char fname[256];

	node=head;
	
	sprintf(fname,"%s_0.html",filename);
	if ((f1=fopen(fname,"w"))==NULL)  {
		fprintf(stderr,"Error opening TOC file\n");
		exit(1);
	}

/* write header */

	fprintf(f1,"<HTML>\n");		/* A html file sig */
	fprintf(f1,"%s\n",HEAD);
	fprintf(f1,"%s%s%s\n",TITLE,"Title etc",ETITLE);
	fprintf(f1,"%s\n",EHEAD);
	fprintf(f1,"%s\n",BODY);

	if (head_count>0) {
		sprintf(fname,"%s_1.html",filename);
		fprintf(f1,"%s",imagelink(fname,
			  NEXT, "BOTTOM", TNEXT));
	}
	fprintf(f1,"%s",imagelink(backlink,TOP, "BOTTOM", TTOP));

	fprintf(f1,"%s",imagelink(WELCOME,HOME, "BOTTOM", THOME));

	fprintf(f1,"%s%s%s%s\n",HEAD1,title,EHEAD1,PARAGRAPH);
	fprintf(f1,"%s\n",RULE);

	for (;node!=NULL;node=node->next) {
	    write_level(f1,node->depth);
	    fprintf(f1,"%s %s\n",
		textlink(node->filename, node->header),
		BREAK);
	}
	
	write_level(f1,1);	/* set level back to 1 */

/* write trailers */

	fprintf(f1,"%s",RULE);

	if (head_count>0) 
		fprintf(f1,"%s",imagelink(fname,
			  NEXT, "BOTTOM", TNEXT));
	fprintf(f1,"%s",imagelink("previous.html",
			  TOP, "BOTTOM", TTOP));

	fprintf(f1,"%s",imagelink(WELCOME,HOME, "BOTTOM", THOME));

	fprintf(f1,"%s%s%s\n",PARAGRAPH,textlink(CRFILE,COPYRIGHT),PARAGRAPH);
	fprintf(f1,"%s\n",textlink(MYLINK, MYTAG));

	fprintf(f1,"%s\n",EBODY);
	fclose(f1);
}

/***
 * write_level()
 *	Indents the TOC line relative to depth of header
 *
 ***/
write_level(f,depth)
FILE *f;
int depth;
{
	int i;
	int static current_depth=1;

	if (current_depth > depth) {
		for (i=depth;i<current_depth;i++)
			fprintf(f,"%s",EINDENT);
		current_depth = depth;
	} else if (current_depth < depth ) {
		for (i=current_depth;i<depth;i++)
			fprintf(f,"%s",INDENT);
		current_depth = depth;
	}
}

char *filter(line)
char *line;
{
	static char new[4*LEN+XTRA];	/* at worst, line can expand 4X */
	int i,j=0;

	for (i=0;i<(4*LEN+XTRA);i++) new[i] = 0;
	i=0;
	while (line[i]!='\0') {
	    switch(line[i]) {
		case '<' : strcat(&new[j],LT);
			   j=j+strlen(LT);
			   break;
		case '>' : strcat(&new[j],GT);
			   j=j+strlen(GT);
			   break;
		case '&' : strcat(&new[j],AMP);
			   j=j+strlen(AMP);
			   break;
		default :  new[j++] = line[i];
	    }
	    i++;
	}
	return(new);
}
/***
 * fix_line()
 *	extract spaces and numbers from a line if we are 
 *	reading the table of contents, otherwise line stays
 *	as is.
 *
 ***/
struct NODE *add_node(s)
char *s;
{

	int i,j;
	char *p, *tmp, str[LEN];
	struct NODE *n, *lnode;

	lnode = (struct NODE *) malloc(sizeof(struct NODE));

	sprintf(str,"%s_%d.html",filename, file_count++);
	lnode->filename=strdup(str);

	lnode->header = strdup(s);
	lnode->next = NULL;
	lnode->depth=1;		/* always 1 */	
	lnode->part=1;

	return(lnode);
}
/***
 * fix_line()
 *	Extract leading white spaces and trailing white spaces
 *	and page numbers.
 ***/
char *fix_line(line)
char *line;
{

	int i;
	char *tmp, *p;

	i=0;
	tmp = strdup(line);

	p=tmp;
	tmp=tmp+strlen(tmp)-1;

/* extract trailing crap of string, eg. pagenumbers */

	while (isspace(*tmp) || isdigit(*tmp)) 
		tmp--;	
	*(tmp+1)='\0';
	tmp=p;

/* extract leading white spaces */

	while(isspace(*tmp)) tmp++;

	return(tmp);
}

/***
 * check_paragraph()
 *	Scans the linked list of headers for duplicates. If found 
 *	this then assumes the start of a new paragraph/section.
 * 	It then sets 'end_toc' to 1 indicating the end of the
 *	table of contents.
 ***/
int check_paragraph(line)
{

	int i;
	char *p, *tmp;
	struct NODE *n;

	i=0;

	n = head;
	while (n) {
		if (strcmp(n->header,line)==0) {
			if (!end_toc) {
				end_toc=1;
				node=head;	/* reset list */
			}
			i=1;
			break;
		}
	        n=n->next;
	}
	
	return(i);
}


struct ENTRY  *readcfg(fname)
char *fname;
{

	FILE *f;
	char line[LEN], *p;
	int i,j;
	struct ENTRY *head=NULL, *entry;

	if ((f=fopen(fname,"r"))==NULL)
		return(0);

	while (fgets(line,LEN,f)!=NULL) {
		i=0;
		while (isspace(line[i]) && line[i]!=0) i++;
		if (line[i]=='#' || strlen(line)==i)
			continue;
		j=i;
		while (!isspace(line[j])) j++;
		
		if (!head) { 
		    entry = (struct ENTRY *) malloc(sizeof(struct ENTRY));
		    head=entry;
		}

		entry->var = (char *) malloc(j+1-i);
		strncpy(entry->var,&line[i],j-i);
		i=j;
		while ((isspace(line[i])) || (line[i]=='=') &&
			line[i]!=0) i++;

		if (line[i]=='\0') 
			entry->val=NULL;
		else { 
			if (line[i]!='"')
				entry->val=NULL;	
			i++; j=i;

			while (line[j]!='"' && line[j]!='\0') j++;
			if (line[j]!='\0') {	
				entry->val = (char *) malloc(j+1-i);
				strncpy(entry->val,&line[i],j-i);
			}
		}
		
		entry->next= (struct ENTRY *) malloc(sizeof(struct ENTRY));
		entry=entry->next;
		entry->next=NULL;			
	}
	return(head);
}	

/***
 * get_cfg()
 *	Returns the string associated with str
 *
 ***/
char *get_cfg(list,str)
struct ENTRY *list;
char *str;
{

	struct ENTRY *tmp;
	char *val=NULL;

	tmp = list;
	while (list) { 
		if ((list->var) && (strcmp(list->var, str)==0))  {
			if (list->val)
				val = (char *) strdup(list->val);	
			break;
		}
		list=list->next;
	}
	list=tmp;
	return(val); 
}

/***
 * init_vars()
 *	This initialiases the paramters from the configuration file
 *
 ***/
init_vars()
{
	struct ENTRY *list;

	list = readcfg(CFGFILE);

	WELCOME = get_cfg(list,"WELCOME");
	COPYRIGHT = get_cfg(list,"COPYRIGHT");
	CRFILE = get_cfg(list,"CRFILE");
	BODY = get_cfg(list,"BODY");
	EBODY = get_cfg(list,"EBODY");
	RULE = get_cfg(list,"RULE");
	HEAD = get_cfg(list,"HEAD");
	EHEAD = get_cfg(list,"EHEAD");
	TITLE = get_cfg(list,"TITLE");
	ETITLE = get_cfg(list,"ETITLE");
	HEAD1 = get_cfg(list,"HEAD1");
	EHEAD1 = get_cfg(list,"EHEAD1");
	HEAD2 = get_cfg(list,"HEAD2");	
	EHEAD2 = get_cfg(list,"EHEAD2");
	PARAGRAPH = get_cfg(list,"PARAGRAPH");
	MARK = get_cfg(list,"MARK");
	EMARK = get_cfg(list,"EMARK");
	INDENT = get_cfg(list,"INDENT");
	EINDENT = get_cfg(list,"EINDENT");
	BREAK = get_cfg(list,"BREAK");

	NEXT = get_cfg(list,"NEXT");
	PREVIOUS = get_cfg(list,"PREVIOUS");
	TOP = get_cfg(list,"TOP");
	HOME = get_cfg(list,"HOME");

	TNEXT = get_cfg(list,"TNEXT");
	TPREVIOUS = get_cfg(list,"TPREVIOUS");
	TTOP = get_cfg(list,"TTOP");
	THOME = get_cfg(list,"THOME");

	return(0);
}

/* new header line */
usage(myname)
char *myname;
{
	fprintf(stderr,"%s -f <text file>\n\t\
[-t <title>] - The main title of this doc.\n\t\
[-p <back link>] - The filename of the back link.\n\t\
[-k] - Keep text format as is, no joining lines\n\t\
[-g] - Generate Table of Contents.\n",myname);

	exit(1);
}
